// RFM69.c 25.02.21, Harald Peters
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <string.h>
#include "RFM69.h"
// 21.05.18: 8MHz

struct tcnt fcnt =
{
// Timer mit 1:1024;   7812Hz,128us; max 32.7ms
// 8MHz/1024 128us  
  .tcntScan = 10, .tcntPause = 150, 
};

static __inline__ void InsertEinsBit(void)
{
 FBlock.BitFeld[FBlock.BitZ] = 1; FBlock.BitZ++;
}

void OOKFlanke(void) // in Hauptschleife aufgerufen, wenn RFM69Stat&0x02
{
 if (TCCR0B && TCNT0<fcnt.tcntScan) return;

 if (TCNT0>fcnt.tcntPause && FBlock.Bit2Start==0) 
 {
  RFM69Stat &= ~0x04;
  if (FBlock.BitZ>40) {FBlock.Bit2Start = FBlock.BitZ; InsertEinsBit();}
  TCNT0 = 0; return;
 } 

 if (RFM69Stat&0x02)
 {
  RFM69Stat &= ~0x02;
  if ((FBlock.BitZ==7) && (PIND&(1<<PD2)) ) {FBlock.x = 1;InsertEinsBit();}

  if (PIND&(1<<PD2)) FBlock.BitFeld[FBlock.BitZ] = 0;// Pegel abtasten
  else               FBlock.BitFeld[FBlock.BitZ] = 1;

 
  FBlock.BitZ++;
  if (TCNT0>20) FBlock.Fehl |= 0x01;
  TCNT0 = 0;
 }
}


ISR(INT0_vect)
{
 TCNT0 = 0;
 TCCR0B  = tccr0; // Timer starten; Ende durch Timerberlauf
 RFM69Stat |= 0x06;
}


uint8_t CRC8( uint8_t *data, uint8_t len) 
{
 uint8_t crc = 0xFF;

 while (len--) 
 {
  uint8_t inbyte = *data++;
  for (uint8_t i=8; i; i--) 
  {
   uint8_t Msb = (crc ^ inbyte) & 0x80; 
   crc <<= 1; 
   if (Msb) crc ^= 0x31;
   inbyte <<= 1; 
  }
 }
 return crc;
}


void FAnalyse(void)
{
 uint8_t BitMuster = 0x80, hByte = 0, ByteIndex = 0, BitIndex = 0, crc;

 for(uint8_t i=0; i<FBlock.BitZ; i++)
 {
   if (i && i==FBlock.Bit2Start) {ByteIndex = 1; BitIndex = 0;}
   if (FBlock.BitFeld[i]) hByte |= BitMuster;
   BitMuster >>= 1;
   if (BitMuster==0) 
   {
    if (BitIndex>=TBMAX) FBlock.Fehl |= 0x02;
    FBlock.TelBytes[ByteIndex][BitIndex++] = hByte;

    BitMuster = 0x80;
    hByte = 0;
   }
 } // for(uint8_t i=0; i<FBlock.BitZ; i++}

 crc = CRC8(&FBlock.TelBytes[0][0],10);
 if (crc!=0) FBlock.Fehl |= 0x10;
 FBlock.x = crc;
 if (FBlock.BitZ<100) return;
 if (CRC8(FBlock.TelBytes[1],10) ) FBlock.Fehl |= 0x20;
}


uint8_t RFM69Spi(uint8_t Adr, uint8_t Inh) // 
{
 RFM69RESET;
 SPDR = Adr;
 while ( (SPSR&(1<<SPIF))==0 ) {}
 SPDR = Inh;
 while ( (SPSR&(1<<SPIF))==0 ) {}
 RFM69SET;
 return SPDR;
}

void RFM69WrReg(uint8_t Adr, uint8_t Inh)
{
 RFM69RESET;
 SPDR = (Adr | 0x80); while ( (SPSR&(1<<SPIF))==0 ) {}
 SPDR = (Inh);        while ( (SPSR&(1<<SPIF))==0 ) {} 
 PORTB |=  (1<<PB2);
}

uint8_t RFM69RdReg(uint8_t Adr)
{
 RFM69RESET;
 SPDR = (Adr & 0x7F);  while ( (SPSR&(1<<SPIF))==0 ) {}
 SPDR = 0;             while ( (SPSR&(1<<SPIF))==0 ) {}
 PORTB |=  (1<<PB2);
 return SPDR;
}


uint16_t RFM69RdRegPaar(uint8_t Reg) 
{
 return (RFM69RdReg(Reg)<<8) | RFM69RdReg(Reg+1);
}

void RFM69WrRegPaar(uint8_t Reg, uint16_t Inh) 
{
 RFM69WrReg(Reg,Inh>>8);
 RFM69WrReg(Reg+1,Inh);
}


uint16_t RFM69BitRate(uint16_t BRate) // in Hz
{
 uint16_t BRateH = 32E6/BRate; 
 if (BRate>0 && BRate<0xFFFF)
 {
  RFM69WrReg(0x03,BRateH>>8);
  RFM69WrReg(0x04,BRateH);
 }
 BRateH = (RFM69RdReg(0x03)<<8) + RFM69RdReg(0x04);
 return 32E6/BRateH;
}

uint32_t RFM69Freq(uint32_t Freq) // 43392 -> 433.92MHz
{
 uint8_t  Reg = 0x07;
 uint32_t FreqH;
 if (Freq>0 && Freq<0x100000) 
 {
  FreqH = (Freq*16384UL)/100; // 2^19 = 524288
  RFM69WrReg(Reg++,FreqH>>16);
  RFM69WrReg(Reg++,FreqH>>8);
  RFM69WrReg(Reg  ,FreqH);
 }
 Reg = 0x07;
 FreqH =          ((uint32_t) RFM69RdReg(Reg++)) << 16;
 FreqH = FreqH + (((uint32_t) RFM69RdReg(Reg++)) <<  8);
 FreqH = FreqH +              RFM69RdReg(Reg);
 return (FreqH  *100UL) / 16384; // 43392 -> 433.92MHz; 2^19 = 524288
}



uint8_t RFM69InitOOK(void) // 
{
 PORTB |=   (1<<PB2); // NSS
 SPCR   =  (1<<SPE) | (1<<MSTR);
 SPSR   =  (1<<SPI2X);

 RFM69Vers = RFM69RdReg(0x10);
 if (RFM69Vers==0x24)
 {
  RFM69WrReg(0x26,0x17); // RSSI auf DIO5; Clkout ausgeschaltet
  RFM69WrReg(0x01,0x10); // 01h:10h Receiver mode
  RFM69BitRate(9600);
  RFM69Freq(86830);
  RFM69WrReg(0x1b,0x40); // 1bh:40h  Schwelle = Peak
  RFM69WrReg(0x1d,100);  // 1dh:60  Schwelle = 60; hhere Werte=niedrigere Empfindlichkeit

  RFM69WrReg(0x19,0x49); // 19h:49h; Bandbreite Bw=32MHz/(20*2^4)=100KHz
  RFM69WrReg(0x02,0x68); // 02:68h OOk continues ohne Bitsychr.
  RFM69Stat |= 0x01;
 }
 return RFM69Vers;
}



ISR(TIMER0_OVF_vect) // fertig
{
 TCCR0B  = 0; // Timer aus
 EIMSK  &= ~(1<<INT0);
 RFM69Stat |= 0x20;
}


void StartStoppFunk(uint8_t Art)
{
 TIFR0  |= (1<<TOV0);  // evtl. Int-Flag lschen
 EIFR   |= (1<<INTF0);

 if (Art==0) // Stopp
 {
  EIMSK  &= ~(1<<INT0);
  TCCR0B = 0;
  return;
 }

 if (Art) // 1. Mal
 {
  if (Art==1) EICRA  = (1<<ISC01);            // neg. INT0-Flanke
  if (Art==2) EICRA  = (1<<ISC01)|(1<<ISC00); // pos. INT0-Flanke
  if (Art==3) EICRA  = (1<<ISC00);            // pos/neg. INT0-Flanke
  tccr0  = (1<<CS01)|(1<<CS00);    // Timer mit 1: 64;  125  KHz,  8us; max 2.48ms
  tccr0  = (1<<CS02)|(0<<CS00);    // Timer mit 1: 256; 31.25KHz, 32us; max 8.19ms
  tccr0  = (1<<CS02)|(1<<CS00);    // Timer mit 1:1024;   7812Hz,128us; max 32.7ms
  TIMSK0 = (1<<TOIE0);
 }

 memset(&FBlock,0,sizeof(FBlock));
 EIMSK |= (1<<INT0); 
}


